/*
 * Copyright CERN 2021
 * Author: Juan David Gonzalez Cobas <dcobas _at_ cern.ch>
 * SPDX-License-Identifier: LGPL-3.0-or-later
 */

#ifndef FSI_LIB_H_
#define FSI_LIB_H_

#ifdef __cplusplus
extern "C" {
#endif

#include <stdint.h>

/**
 * general questions/open ends --- look fast for 'tbd' or FIXME in the
 * code to quickly identify loose ends
 *
 * - some params like frequencies and distances might better be passed
 *   as large integers/indexes in a conventionally bound array between
 *   e.g. 1525-1630nm - tbd
 * - channel_ids are provisionally defined as integers. A more
 *   sophisticated addressing (e.g. HW or geographical) might be defined
 *   lated - tbd
 * - channels will be distance or Bragg, and so configured till next
 *   call to the configuration (INPUT) function
 * - it is assumed that a channel does not mix distance/Bragg
 *   measurements - am I wrong?
 * - the peak set requested in a channel is assumed to have different
 *   fit params per peak. If a channel has all its peaks of the same
 *   type, this can be size-optimized - tbd
 */

/** types of channel inputs: distance meas. or Bragg */
enum fsi_channel_type {
	FSI_CH_NONE = 0,
	FSI_CH_DISTANCE = 1,	/**< configure as distance meas. channel */
	FSI_CH_BRAGG,		/**< configure as Bragg channel */
	FSI_CH_GAS_CELL,	/**< gas cell channel */
	FSI_CH_INTERF,		/**< reference interferometer channel */
};

/** types of functional forms to fit the peaks found */
enum fsi_peak_type {
	FSI_PEAK_NONE = 0,
	FSI_PEAK_LORENTZ = 1,	/**< Cauchy-Lorentz 1/(1+xˆ2) distribution */
	FSI_PEAK_GAUSS,		/**< Gaussian (normal) distribution */
	FSI_PEAK_SINC,		/**< Sinc (sin(x)/x)) distribution */
	FSI_PEAK_MOFFAT,	/**< Moffat (power of Cauchy pdf) distribution */
	FSI_PEAK_BRAGG,		/**< Bragg-type peak */
			/* FIXME:should Bragg type be set apart? tbd */
};
/** types of raw buffer requested by expert diagnostic functions */
enum fsi_expert_mode {
    FSI_EXPERT_NONE = 0,
    FSI_EXPERT_LIN_FFT,		/**< buffer of linearized FFT data */
    FSI_EXPERT_LIN_DATA,	/**< does this mean linearized time-domain data? tbd */
    FSI_EXPERT_OSCILLOSCOPE,	/**< oscilloscope mode - non-linearized raw signal */
    FSI_EXPERT_GAS_CELL,	/**< linearized input data buffer from gas cell */
    FSI_EXPERT_INTERF,		/**< linearized input data buffer from reference interferometer */
    FSI_EXPERT_RAW_GAS_CELL,	/**< non-linearized, raw input data buffer from gas cell */
    FSI_EXPERT_RAW_INTERF,	/**< non-linearized, raw input data buffer from reference interferometer */
    FSI_EXPERT_OSC_BRAGG,	/**< tbd - no different from FSI_EXPERT_OSCILLOSCOPE? */
};

/** biggest number of model parameters among different peak types */
#define FSI_MAX_MODEL_PARAMS	4
/** max number of peaks to find in a channel sample */
#define FSI_MAX_PEAKS		5
/** length of return peak data */
#define	FSI_PEAK_LENGTH		200
/** size of sample vectors - this should not be hardcoded - tbd */
#define	FSI_SAMPLE_SIZE		1024

/* library error codes */
#define	__FSI_ERR_START	1024

/** peak finder/fitting configuration for an individual peak
 *
 * A channel will be configured with up to FSI_MAX_PEAKS of these to
 * find in its spectrum
 */
struct fsi_peak_config {
	double	frequency;			/**< estimated frequency where peak fits */
	double	range1, range2;			/**< estimated interval of peak range */
	double	channel_gain;			/**< FIXME: should be per-channel, not per-peak? */
	enum fsi_peak_type type;		/**< shape of peak to fit */
};

/** peak finder/fitting result for an individual peak
 *
 * A channel measurement result will be constituted by up to
 * FSI_MAX_PEAKS of these
 */
struct fsi_peak_result {
	double		frequency;		/**< center of peak fitted to data */
	double		amplitude;		/**< peak width derived from parameters */
	double		snr;			/**< peak signal-to-noise ratio (? tbd) */
	enum fsi_peak_type
			type;			/**< shape of peak fit to data */
	double		fit_params[FSI_MAX_MODEL_PARAMS];
						/**< goodness-of-fit params, model-dependent */
	uint16_t	samples[FSI_PEAK_LENGTH];	/**< actual peak samples */

};

/**
 * channel configuration - this will set up a channel's data processing
 * pipeline for the next acq cycle. Up to FSI_MAX_PEAKS will be
 * requested
 */
struct fsi_channel_config {
	unsigned int	channel_id;		/**< channel to be configured */
	enum fsi_channel_type	type;		/**< type of channel (distance/Bragg) */
	double		channel_gain;		/**< channel gain FIXME: see fsi_peak_request */
	int		npeaks;			/**< number of valid peaks to be requested in peaks array */
	struct fsi_peak_config
			peaks[FSI_MAX_PEAKS];	/**< peak requests 
							FIXME: should peak params be identical for all of them?
							If so, this needs not be an array and npeaks is sufficient */
    };

/**
 * channel measurement results - the result of acq data processing for
 * an individual channel is returned here, with npeaks (up to 
 * FSI_MAX_PEAKS) peaks given in the case by their fit results (type of
 * function and its params, and an array of FSI_PEAK_LENGTH samples)
 */
struct fsi_channel_measurement {
	unsigned int		channel_id;		/**< channel id of this measurement */
	enum fsi_channel_type	type;			/**< type of channel (distance/Bragg) */
	int			npeaks;			/**< number of valid entries in peaks array */
	struct fsi_peak_result	peaks[FSI_MAX_PEAKS];	/**< peak measured values */
};

/**
 * this collects full/raw data for an individual channel, as per the
 * expert/diagnostic functions
 */
struct fsi_full_measurement {
	unsigned int		channel_id;		/**< channel id of this measurement */
	enum fsi_channel_type	type;			/**< type of channel (distance/Bragg) */
	enum fsi_expert_mode	mode;			/**< mode of raw data: linearized FFT/data/oscill-raw/gas-cell */
	double			data[FSI_SAMPLE_SIZE];	/**< raw data */
};

/** opaque library handle/context structure */
typedef struct __FSI_state FSI;

/* core API methods */

/** initialize library */
FSI *fsi_lib_init(void);

/** close library */
void fsi_lib_exit(FSI *h);

/** configure a channel type, gain and expected peak params */
int fsi_channel_configure(FSI *h, struct fsi_channel_config *cfg);

/** get a channel last configured params */
int fsi_get_channel_configuration(FSI *h, unsigned int channel_id, struct fsi_channel_config *cfg);

/** bulk configure a channel array type, gain and expected peak params */
int fsi_channel_array_configure(FSI *h, struct fsi_channel_config cfg[], int nchannels);

/** receive the whole set of distance/bragg measurements from all channels */
int fsi_receive_measurements(FSI *h, struct fsi_channel_measurement meas[], int *nchannels);

/** receive the whole set of distance/bragg measurements from all channels with a timeout */
int fsi_receive_measurements_timemout(FSI *h, struct fsi_channel_measurement meas[], int *nchannels,
					unsigned int timeout);

/** receive raw measurements of channel_id */
int fsi_get_full_samples(FSI *h, int channel_id, enum fsi_expert_mode mode, struct fsi_full_measurement *meas);

/* auxiliary helper methods */
int fsi_get_gas_cell_id(FSI *h);
int fsi_get_interferometer_id(FSI *h);

#ifdef __cplusplus
}
#endif

#endif /* FSI_LIB_H_ */
